/* wol_srv.c */

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/wait.h>
#include <syslog.h>
#include <time.h>
#include <netdb.h>
#include <errno.h>
#include <stdio.h>
#include <unistd.h> /* close() */
#include <stdlib.h>
#include <string.h> /* memset() */
#include <dirent.h>
#include <ctype.h>
#include <fcntl.h>
#include <sys/prctl.h>

#include "wol_srv.h" 

#define LOCAL_SERVER_PORT 7
#define MAX_MSG 100
#define MAX_LINE 2000
#define BUF_SIZE 100


void errExit(int daemon_flag, int exitcode, const char *format, ...)
{
	va_list argList;

	va_start(argList, format);
	if(daemon_flag)
	{
		syslog(LOG_ERR, format,argList);
	}
	else
	{
		printf(format,argList);
	}
	va_end(argList);
	exit(exitcode);
}


void pidfile_create(const char *pidFile, pid_t pid)
{
	char buf[20];
	struct flock lock;
	int fd, value;

	if(*pidFile=='\0')
	{
		// pidfile not defined 
		// return without action
		return;
	}

	if((fd = open(pidFile, O_WRONLY | O_CREAT, (S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH))) < 0) 
	{
		printf("Cannot open pidfile [%s]\n",pidFile);
		exit(EXIT_FAILURE);
	}
	lock.l_type = F_WRLCK;
	lock.l_start = 0;
	lock.l_whence = SEEK_SET;
	lock.l_len = 0;

	if (fcntl(fd, F_SETLK, &lock) < 0)
	{
		if (errno != EACCES && errno != EAGAIN)
			printf("Cannot lock pidfile [%s]\n",pidFile);
		else
			printf("process is already running\n");
		close(fd);
		exit(EXIT_FAILURE);
	}
	if (ftruncate(fd, 0) < 0) 
	{
		printf("Cannot truncate pidfile [%s]\n",pidFile);
		close(fd);
		exit(EXIT_FAILURE);
	}
	memset(buf, 0, sizeof(buf));
	snprintf(buf, sizeof(buf)-1, "%u", pid);
	if (write(fd, buf, strlen(buf)) != strlen(buf))
	{
		printf("Cannot write pid to pidfile [%s]\n",pidFile);
		close(fd);
		exit(EXIT_FAILURE);
	}
	if ((value = fcntl(fd, F_GETFD, 0)) < 0)
	{
		printf("Cannot get close-on-exec flag from pidfile [%s]\n", pidFile);
		close(fd);
		exit(EXIT_FAILURE);
	}
	value |= FD_CLOEXEC;
	if (fcntl(fd, F_SETFD, value) < 0)
	{
		printf("Cannot set close-on-exec flag from pidfile [%s]\n", pidFile);
		close(fd);
		exit(EXIT_FAILURE);
	}

	// leave with opend file to lock
}

int search_proc(char *search)
{
	DIR *dirp;
	struct dirent *dp;
	char path[PATH_MAX];
	char line[MAX_LINE], cmd[MAX_LINE];
	FILE *fp;
	char *p;
	uid_t uid, checkedUid;
	int gotName, gotUid;
	size_t len,i;

	// only open proc if search-string is set
	if(search[0]=='\0')
	{
		return 0;
	}

	dirp = opendir("/proc");
	if (dirp == NULL)
	{
		// error oben /proc
		return -1;
	}

	// Scan entries under /proc directory 

	for (;;) 
	{
		errno = 0;			  /* To distinguish error from end-of-directory */
		dp = readdir(dirp);
		if (dp == NULL) 
		{
			if (errno != 0)
			{
				// readerror for /proc
				return -1;
			}
			else
			{
				break;
			}
		}

		// look only for dir-entries startet with numbers (PID's)

		if (dp->d_type != DT_DIR || !isdigit((unsigned char) dp->d_name[0]))
		{
			continue;
		}

		snprintf(path, PATH_MAX, "/proc/%s/status", dp->d_name);

		fp = fopen(path, "r");
		if (fp == NULL)
		{
			continue;		// Ignore errors: fopen() might fail if process has just terminated
		}

		cmd[0]='\0';
		while (!gotName)
		{
			if (fgets(line, MAX_LINE, fp) == NULL)
			{
				break;
			}

			// search for "Name:"
			if (strncmp(line, "Name:", 5) == 0) 
			{
				// skip space after Name
				for (p = line + 5; *p != '\0' && isspace((unsigned char) *p); )
					p++;
				// copy rest of line to cmd
				strncpy(cmd, p, MAX_LINE - 1);
				cmd[MAX_LINE -1] = '\0';		/* Ensure null-terminated */
				break;
			}
		}

		fclose(fp);

		snprintf(path, PATH_MAX, "/proc/%s/cmdline", dp->d_name);

		fp = fopen(path, "r");
		if (fp != NULL)
		{
			len=fread(line, 1, MAX_LINE-2, fp);
			// if cmdline have data
			if (len)
			{
				for(i=0;i<len;i++)
				{
					if(line[i]=='\0')
					{
						line[i]=' ';
					}
				}
				line[len]='\n';
				line[len+1]='\0';
				strcpy(cmd,line);
				cmd[MAX_LINE -1] = '\0';		/* Ensure null-terminated */
			}
			fclose(fp);

		}
//		printf("%5s %s", dp->d_name, cmd);

		// is search-string in cmd?
		if(strstr(cmd,search)!=NULL)
		{
			// we found the string
			return 1;
		}
	}
	return 0;
}

cfg_t *parse_conf(const char *filename)
{
	cfg_opt_t srv_opts[] = 
	{
		CFG_INT("timeout", 0, CFGF_NONE),
		CFG_STR("mac", "", CFGF_NONE),
		CFG_STR("ignore-start", "", CFGF_NONE),
		CFG_STR("start", "", CFGF_NONE),
		CFG_END()
	};

	cfg_opt_t opts[] = 
	{
		CFG_SEC("srv", srv_opts, CFGF_MULTI | CFGF_TITLE),
		CFG_BOOL("daemon", cfg_true, CFGF_NONE),
		CFG_BOOL("only-one-packet", cfg_true, CFGF_NONE),
		CFG_BOOL("debug", cfg_false, CFGF_NONE),
		CFG_INT_LIST("udp-ports", "{7, 9}", CFGF_NONE),
		CFG_FUNC("include-file", cfg_include),
		CFG_STR("pidfile", "", CFGF_NONE),
		CFG_END()
	};

	cfg_t *cfg = cfg_init(opts, CFGF_NONE);

	switch(cfg_parse(cfg, filename))
	{
		case CFG_FILE_ERROR:
			printf("warning: configuration file '%s' could not be read: %s\n",filename, strerror(errno));
			printf("continuing with default values...\n\n");
		case CFG_SUCCESS:
			break;
		case CFG_PARSE_ERROR:
			return 0;
	}

	return cfg;
}

void conv_msg(char *msg, int len, char *msg3)
{
	int i;
	for(i=0; i<len; i++)
	{
		sprintf(msg3,"%02x:",(unsigned char)*msg);
		msg3+=3;
		msg++;
	}
	*msg3='\0';
}

int main(int argc, char *argv[]) 
{
	int i;
	int status = 0;
//	int sd, rc, n, cliLen;
	cfg_bool_t debug_flag,daemon_flag;
//	struct sockaddr_in cliAddr, servAddr;
//	char msg[MAX_MSG];
//	int opt = 1;
	time_t to[MAX_SRV];
	pid_t pid;
	pid_t sid;
	pid_t child_pid, wpid;

	cfg_t *cfg = parse_conf(argc > 1 ? argv[1] : "wol_srv.conf");

	if(cfg)
	{
		// init timeout
		for(i=0;i<MAX_SRV;i++)
		{
			to[i]=0;
		}

		daemon_flag=cfg_getbool(cfg, "daemon");
		debug_flag=cfg_getbool(cfg, "debug");

		if(cfg_size(cfg, "srv")>=MAX_SRV)
		{
			errExit(daemon_flag, EXIT_FAILURE, "%s Error number of servers to big (max %d) %d \n",DAEMON_NAME,MAX_SRV,cfg_size(cfg, "srv"));
		}

		if(daemon_flag)
		{
			/* Clone ourselves to make a child */  
			pid = fork(); 
 
			/* If the pid is less than zero, something went wrong when forking */
			if (pid < 0) 
			{
				syslog(LOG_ERR, "%s Error starting daemon %d \n",DAEMON_NAME,pid); 
				exit(EXIT_FAILURE);
			}
 
			/* If the pid we got back was greater than zero, then the clone was successful and we are the parent. */
			if (pid > 0) 
			{
				exit(EXIT_SUCCESS);
			}
			/* Set the umask to zero */
			umask(0);

			/* Sends a message to the syslog daemon */
			syslog(LOG_NOTICE, "%s Successfully started daemon\n",DAEMON_NAME); 
 
			/* Try to create our own process group */
			sid = setsid();
			if (sid < 0) 
			{
				errExit(daemon_flag, EXIT_FAILURE, "%s Could not create process group\n",DAEMON_NAME);
			}
			/* Change the current working directory */
			if ((chdir("/")) < 0) 
			{
				errExit(daemon_flag, EXIT_FAILURE, "%s Could not change working directory to /\n",DAEMON_NAME);
			}
			/* Close the standard file descriptors */
			close(STDIN_FILENO);
			close(STDOUT_FILENO);
			close(STDERR_FILENO);
		}
 
		/* If execution reaches this point we are the child of daemon == daemon self */

		pidfile_create(cfg_getstr(cfg, "pidfile"), getpid());
		// report all seen configs for server
		if(debug_flag && !daemon_flag)
		{
			for(i = 0; i < cfg_size(cfg, "srv"); i++)
			{
				cfg_t *srv = cfg_getnsec(cfg, "srv", i);
				printf("Starting: %s\n",cfg_getstr(srv, "start"));
				printf("	ignore if process running like: %s\n",cfg_getstr(srv, "ignore-start"));
				printf("	timewindow: %ld\n",cfg_getint(srv, "timeout"));
				printf("	if mac is: %s\n\n",cfg_getstr(srv, "mac"));
			}
		}

		// start all childs for every port to watch
		for(i = 0; i < cfg_size(cfg, "udp-ports"); i++)
		{
			if ((child_pid = fork()) == 0)
			{
				int sd, rc, n, cliLen,k;
				int opt = 1;
				int last_sport=0;
				struct sockaddr_in cliAddr, servAddr;
				char msg[MAX_MSG];
				char msg3[MAX_MSG*3];
				int lport=cfg_getnint(cfg, "udp-ports", i);

			        prctl(PR_SET_PDEATHSIG, SIGTERM);

				if(daemon_flag)
				{
					syslog(LOG_INFO, "%s open child-process (pid %d) for udp-port %ld\n",DAEMON_NAME,getpid(),cfg_getnint(cfg, "udp-ports", i));
				}
				else if(debug_flag)
				{
					printf("open child-process (pid %d) for port udp-ports #%d: %ld\n",getpid(),i+1,cfg_getnint(cfg, "udp-ports", i));
				}
				/* socket creation */
				sd=socket(AF_INET, SOCK_DGRAM, 0);
				if(sd<0) 
				{
					errExit(daemon_flag, EXIT_FAILURE, "%s cannot open socket\n",DAEMON_NAME);
				}

				setsockopt(sd, SOL_SOCKET, SO_REUSEADDR, &opt, sizeof(opt));

				/* bind local server port */
				servAddr.sin_family = AF_INET;
				servAddr.sin_addr.s_addr = htonl(INADDR_ANY);
				servAddr.sin_port = htons(lport);
				rc = bind (sd, (struct sockaddr *) &servAddr,sizeof(servAddr));
				if(rc<0) 
				{
					errExit(daemon_flag, EXIT_FAILURE, "%s cannot bind port number %d\n",DAEMON_NAME,lport);
				}

				printf("%s: waiting for data on port UDP %u\n", DAEMON_NAME,lport);

				/* server infinite loop */
				while(1) 
				{
					int sport;
					/* init buffer */
					memset(msg,0x0,MAX_MSG);

					/* receive message */
					cliLen = sizeof(cliAddr);
					n = recvfrom(sd, msg, MAX_MSG-2, 0, (struct sockaddr *) &cliAddr, &cliLen);

					sport=ntohs(cliAddr.sin_port);

					if(sport!=last_sport)
					{
						if(n<0)
						{
							if(daemon_flag)
							{
								syslog(LOG_WARNING, "%s cannot receive data\n",DAEMON_NAME);
							}
							else
							{
								printf("%s: cannot receive data \n",DAEMON_NAME);
							}
							continue;
						}
						else
						{
							char *mac;
							char *st;
							char *ist;
							const char *ti;
							conv_msg(msg, n, msg3);

							if(debug_flag && !daemon_flag)
							{
								printf("%s: receive message on port %d from %s, message %s\n",DAEMON_NAME,lport,inet_ntoa(cliAddr.sin_addr),msg3);
							}
							for(k = 0; k < cfg_size(cfg, "srv"); k++)
							{
								cfg_t *srv = cfg_getnsec(cfg, "srv", k);
								mac=cfg_getstr(srv, "mac");
								ti=cfg_title(srv);
								
								if(debug_flag && !daemon_flag)
								{
									printf("%s: srv-config %s found, look for mac %s\n",DAEMON_NAME,ti,mac);
								}
								if(strstr(msg3,mac)!=NULL || strstr(msg,mac)!=NULL)
								{
									st=cfg_getstr(srv, "start");
									ist=cfg_getstr(srv, "ignore-start");
									if(debug_flag && !daemon_flag)
									{
										printf("%s: srv-config %s , found mac %s\n",DAEMON_NAME,ti,mac);
									}
									if(time(NULL)>(to[k]+cfg_getint(srv, "timeout")))
									{
										// the packet is in the right time
										to[k]=time(NULL);
										if(search_proc(ist)==0)
										{
											if(daemon_flag)
											{
												syslog(LOG_INFO, "%s found MAC %s on port %d, starting %s\n",DAEMON_NAME,mac,lport,ti);
											}
											else if(debug_flag)
											{
												printf("%s: found MAC %s on port %d, starting %s\n",DAEMON_NAME,mac,lport,ti);
											}
											if(st && *st!='\0')
											{
												system(st);
											}
											else if(debug_flag)
											{
												printf("%s: can not start %s, no start-string\n",DAEMON_NAME,ti);
											}
										}
										else
										{
											if(daemon_flag)
											{
												syslog(LOG_INFO, "%s ignore found MAC %s on port %d, %s already startet\n",DAEMON_NAME,mac,lport,ti);
											}
											else if(debug_flag)
											{
												printf("%s: ignore found MAC %s on port %d, %s already started\n",DAEMON_NAME,mac,lport,ti);
											}
										}
									}
									else
									{
										if(daemon_flag)
										{
											syslog(LOG_INFO, "%s time-ignore UDP for %s\n",DAEMON_NAME,ti);
										}
										else if(debug_flag)
										{
											printf("%s: time-ignore UDP for %s\n",DAEMON_NAME,ti);
										}
									}
								}
							}
						}

						if(cfg_getbool(cfg, "only-one-packet"))
						{
							last_sport=sport;
						}
						else
						{
							last_sport=0;
						}
					}
      
				}/* end of server infinite loop */

				exit(0);
				/*NOTREACHED*/
			}
		}

		while ((wpid = wait(&status)) > 0)
		{
			;
//			printf("Exit status of %d was %d (%s)\n", (int)wpid, status,(status > 0) ? "accept" : "reject");
		}

		cfg_free(cfg);
		return 0;
	}
	return 0;
}

